import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.linear_model import Lasso
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OneHotEncoder, StandardScaler

from scatcluster.helper import COLORS


class ExternalCorrelation:

    def plot_external_correlation(self,
                                  df_predictions: pd.DataFrame,
                                  df_external: pd.DataFrame,
                                  metric_list: list[str],
                                  title: str = None,
                                  **kwargs):

        gs = {'height_ratios': [1] + [0.2] * len(metric_list)}
        fig, axs = plt.subplots(len(metric_list) + 1, 1, figsize=(12, 10), sharex=True, gridspec_kw=gs, **kwargs)

        num_clusters = np.unique(df_predictions['predictions'])
        for clust in num_clusters:
            dt = df_predictions.loc[
                df_predictions['predictions'] == clust,
            ]
            axs[0].vlines(dt['dates'], clust - 0.5, clust + 0.5, color=COLORS[clust])
        axs[0].set_yticks(range(1, df_predictions['predictions'].max() + 1))
        axs[0].set_ylabel('Cluster')

        for metric_enum, metric in enumerate(metric_list):
            metric_enum += 1
            axs[metric_enum].plot(df_external['dates'], df_external[metric])
            axs[metric_enum].set_ylabel(metric, rotation=30)

        _title = '01_pre_truncate' if title is None else title
        plot_name = f'{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_ICA_{self.ica.n_components}_clusters_{num_clusters}_ExternalCorrelation_{_title}'
        plt.suptitle(plot_name)
        plt.savefig(f'{self.data_savepath}figures/{plot_name}.png')

    def merge_same_time_duration(
        self,
        df_predictions: pd.DataFrame,
        df_external: pd.DataFrame,
    ):

        ## Truncate data to the same time duration
        date_start = max(df_external.DATES.min(), df_predictions.DATES.min())
        date_end = min(df_external.DATES.max(), df_predictions.DATES.max())

        df_predictions_same = df_predictions.loc[
            ((df_predictions.DATES >= date_start) & (df_predictions.DATES <= date_end)),
        ]
        df_external_same = df_external.loc[
            ((df_external.DATES >= date_start) & (df_external.DATES <= date_end)),
        ]

        df_merge = pd.merge(left=df_predictions_same[['dates', 'predictions']],
                            left_on='dates',
                            right=df_external_same,
                            right_on='dates',
                            how='left')

        return df_merge

    def interpolate_missing_values(self,
                                   df_merge: pd.DataFrame,
                                   target_cols: list[str] = [
                                       'outTemp', 'outHumidity', 'barometer', 'windSpeed', 'windDir', 'windGust',
                                       'windGustDir', 'rain', 'rainRate', 'dewpoint'
                                   ]):
        df_merge[target_cols] = df_merge[target_cols].interpolate(method='spline',
                                                                  limit_direction='both',
                                                                  order=1,
                                                                  axis=0)
        return df_merge

    def detection_rate(self,
                       df_merge: pd.DataFrame,
                       rolling_window_size: int = 48,
                       plot_detection: bool = True,
                       title: str = None):
        # Use Detection Rate to determine the quality of Fit against Weather Data

        ## Cluster OHE & Smoothing
        cluster_transformer = Pipeline(steps=[('encoder', OneHotEncoder(handle_unknown='ignore'))])
        df_clusters_OHE = cluster_transformer.fit_transform(np.array(df_merge.predictions).reshape(-1, 1))
        df_clusters_OHE = pd.DataFrame(df_clusters_OHE.todense(), columns=range(1, 11))
        df_detection_rate = df_clusters_OHE.rolling(rolling_window_size).mean()[rolling_window_size - 1:]

        if plot_detection:
            fig, axs = plt.subplots(1, 1, figsize=(10, 5), sharex=True)
            num_clusters = np.unique(df_merge['predictions'])
            for clust in num_clusters:
                dt = df_merge.loc[
                    df_merge['predictions'] == clust,
                ]
                axs.vlines(dt['dates'], clust - 0.5, clust + 0.5, color=COLORS[clust])
                axs.plot(df_merge['dates'][rolling_window_size - 1:],
                         df_detection_rate[clust].values + clust - 0.5,
                         'k',
                         linewidth=0.5)
            axs.set_yticks(range(1, df_merge['predictions'].max() + 1))
            axs.set_ylabel('Cluster')

            _title = '03_cluster_detection' if title is None else title
            plot_name = f'{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_ICA_{self.ica.n_components}_clusters_{num_clusters}_ExternalCorrelation_{_title}'
            plt.suptitle(plot_name)
            plt.savefig(f'{self.data_savepath}figures/{plot_name}.png')
            plt.show()

        return df_detection_rate

    def external_data_smoothing_scaling(self,
                                        df_merge: pd.DataFrame,
                                        target_cols: list[str],
                                        rolling_window_size: int = 48,
                                        plot_external_data_smoothing: bool = True,
                                        title: str = None):
        ## Weather Smoothing & Scaling
        df_weather_smoothed = df_merge[target_cols].rolling(rolling_window_size).mean()[rolling_window_size - 1:]
        weather_transformer = Pipeline(steps=[('scaler', StandardScaler())])
        df_weather_smoothed_scaled = pd.DataFrame(weather_transformer.fit_transform(df_weather_smoothed),
                                                  columns=df_weather_smoothed.columns)

        if plot_external_data_smoothing:
            fig, axs = plt.subplots(len(df_weather_smoothed_scaled.columns), 1, figsize=(12, 10), sharex=True)
            for metric_enum, metric in enumerate(df_weather_smoothed_scaled.columns):
                axs[metric_enum].plot(df_merge['dates'], df_merge[metric])
                axs[metric_enum].plot(df_weather_smoothed_scaled['dates'], df_weather_smoothed_scaled[metric], 'k')
                axs[metric_enum].set_ylabel(metric, rotation=30)

            _title = '04_weather_smoothed' if title is None else title
            plot_name = f'{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_ICA_{self.ica.n_components}_clusters_{num_clusters}_ExternalCorrelation_{_title}'
            plt.suptitle(plot_name)
            plt.savefig(f'{self.data_savepath}figures/{plot_name}.png')
            plt.show()

        return df_weather_smoothed_scaled

    def external_correlation(self, df_clusters_smoothed, df_weather_smoothed_scaled, target_cols):
        # REGRESSION
        # LASSO = L1 : Lasso(fit_intercept=False, alpha=0.00001)
        # RIDGE = L2 : Ridge(fit_intercept=False, alpha=0.00001)
        # regression = LinearRegression(fit_intercept=False)

        SCORE = {}
        REGRESSION = {}
        for col in target_cols:
            regression = Lasso(fit_intercept=False, alpha=0.00001)
            regression.fit(df_clusters_smoothed, df_weather_smoothed_scaled[col])
            score = regression.score(df_clusters_smoothed, df_weather_smoothed_scaled[col])
            print(f'>>> {col} - score: {score}')
            REGRESSION[col] = regression
            SCORE[col] = score

        print(f'>> Best performing weather data {[k for k in SCORE.keys() if SCORE[k] == max(SCORE.values())]}')
        print('\n')

        return SCORE, REGRESSION

    def min_max_vector(self, vector):
        x_min = vector.min()
        x_max = vector.max()
        return (vector - x_min) / (x_max - x_min)

    def plot_external_data_predicted_actual(self, df_clusters_smoothed, df_weather_smoothed_scaled, REGRESSION,
                                            target_cols):

        fig, axs = plt.subplots(1, 1, figsize=(10, 6), sharex=True)
        for clust_enum, col in enumerate(target_cols):
            weather_mm = self.min_max_vector(df_weather_smoothed_scaled[col].values)
            prediction_weather = REGRESSION[col].predict(df_clusters_smoothed)
            prediction_mm = self.min_max_vector(prediction_weather)

            axs.plot(weather_mm + clust_enum, 'k', linewidth=0.5)
            axs.plot(prediction_mm + clust_enum, 'r', linewidth=0.5)
        axs.set_yticks(range(len(target_cols)))
        axs.set_yticklabels(target_cols)
        axs.legend(['Actual', 'Predicted'], loc='upper center', ncols=2, bbox_to_anchor=(0.5, -0.05))

        plot_name = f'{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_ICA_{self.ica.n_components}_clusters_{num_clusters}'
        plt.title(f'{plot_name}\nPredicted vs Actual Weather Data')
        plt.savefig(f'{self.data_savepath}figures/{plot_name}_05_pred_actual_weather.png')
        plt.show()

    def plot_detection_actual_prediction(
        self,
        target_cols,
        df_clusters_smoothed,
        df_weather_smoothed_scaled,
        REGRESSION,
    ):
        for col in target_cols:
            weather_mm = self.min_max_vector(df_weather_smoothed_scaled[col].values)
            prediction_weather = REGRESSION[col].predict(df_clusters_smoothed)
            prediction_mm = self.min_max_vector(prediction_weather)

            fig, axs = plt.subplots(1, 1, figsize=(10, 6), sharex=True)
            for clust_enum, clust in enumerate(range(1, 11)):
                axs.plot(df_clusters_smoothed[clust] + clust_enum, 'b', linewidth=0.5)
                axs.plot(weather_mm + clust_enum, 'k', linewidth=0.5)
                axs.plot(prediction_mm + clust_enum, 'r', linewidth=0.5)
            axs.set_yticks(range(10))
            axs.set_yticklabels([x + 1 for x in range(10)])
            axs.legend(['Detection Rate', 'Actual', 'Predicted'],
                       loc='upper center',
                       ncols=3,
                       bbox_to_anchor=(0.5, -0.05))

            plot_name = f'{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_ICA_{self.ica.n_components}_clusters_{num_clusters}'
            plt.title(f'{plot_name}\n{col} - {SCORE[col]}')
            plt.savefig(f'{self.data_savepath}figures/{plot_name}_06_all_overlay_{col}.png')
            plt.show()

    def plot_regression_coefficients_contributions(self, target_cols, REGRESSION):
        fig, ax = plt.subplots(figsize=[10, 5])

        regression_coefficients = np.abs(
            [np.array(np.array(REGRESSION[col].sparse_coef_.todense())[0]) for col in target_cols]).T
        cc = ax.imshow(regression_coefficients, aspect='auto', cmap='RdYlBu_r')

        fig.colorbar(cc, ax=ax, location='right')
        plt.xticks(range(10), target_cols, rotation=90)
        plt.xlabel('Weather data')
        plt.yticks(range(10))
        plt.ylabel('Cluster')

        plot_name = f'{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_ICA_{self.ica.n_components}_clusters_{num_clusters}'
        plt.title(f'{plot_name}\nRegression Coefficient Contribution of Weather Data')
        plt.savefig(f'{self.data_savepath}figures/{plot_name}_07_Reg_Coeff.png')
        plt.show()

    def plot_regression_coefficients(self, df_clusters_smoothed, df_weather_smoothed_scaled, REGRESSION, SCORE,
                                     target_cols):
        fig, axs = plt.subplots(5, 2, figsize=(8, 8), sharex=True, sharey=True)
        ROW = 0
        COL = -1
        for col_enum, col in enumerate(target_cols):
            weather_mm = self.min_max_vector(df_weather_smoothed_scaled[col].values)
            prediction_weather = REGRESSION[col].predict(df_clusters_smoothed)
            prediction_mm = self.min_max_vector(prediction_weather)

            COL += 1
            if col_enum % 2 == 0 and col_enum > 0:
                ROW += 1
                COL = 0

            axs[ROW, COL].scatter(weather_mm, prediction_mm, s=0.2)
            axs[ROW, COL].set_title(f'{col}-{round(SCORE[col],4)}')

            if COL == 0:
                axs[ROW, COL].set_ylabel('Weather Predicted')
            if ROW == 4:
                axs[ROW, COL].set_xlabel('Weather Actual')

        plot_name = f'{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_ICA_{self.ica.n_components}_clusters_{num_clusters}'
        plt.suptitle(f'{plot_name}\nCorrelation of Weather data')
        plt.savefig(f'{self.data_savepath}figures/{plot_name}_08_weather_corr.png')
        plt.show()



import os
import pickle
from glob import glob
from typing import Optional

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import dates as mdates
from obspy.core import UTCDateTime
from scipy.signal import medfilt
from sklearn.decomposition import FastICA
from sklearn.metrics import explained_variance_score, mean_squared_error
from sklearn.preprocessing import RobustScaler

from scatcluster.helper import demad, list_of_strings


def round_nearest(x, a):
    return round(x / a) * a


class ICA:

    def _get_index_from_UTC_timestamp(self, utc_timestamp: str):
        df_time = pd.DataFrame({'time': self.data_times})
        return max(df_time.loc[
            df_time['time'] < mdates.date2num(UTCDateTime(utc_timestamp)),
        ].index)

    def process_ICA_single(self,
                           num_ICA: int,
                           return_data: bool = False,
                           exclude_timestamps: Optional[list[str]] = None,
                           exclude_timestamps_skip: int = 5,
                           **kwargs):
        """Process the data for a single run of ICA. This will reduce to a dataset of dimension num_ICA.

        Args:
            num_ICA (int): The number of dimensions to be reduced to.
            return_data (bool, optional): Toggle to determine if the data should be returned as part of the function call. Defaults to False.

        Returns:
            _type_: _description_
        """
        print(f'Performing ICA for {num_ICA} ICAs')
        ica_model_path = f'{self.data_savepath}ICA/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_dimension_{num_ICA}.pickle'
        # check if reduction exists already
        if (os.path.exists(ica_model_path) and self.ica_overwrite_previous_models is False):
            print('  Using pre-calculated model')
            model = pickle.load(open(ica_model_path, 'rb'))
        else:
            # else fit
            kwargs['max_iter'] = 1000 if kwargs.get('max_iter') is None else kwargs.get('max_iter')
            kwargs['whiten'] = 'unit-variance' if kwargs.get('whiten') is None else kwargs.get('whiten')
            model = FastICA(n_components=num_ICA, random_state=42, **kwargs)

        # SCALING
        print('RobustScaling of data before fitting ICA')
        # Exclude any timestamps
        scat_coeff_data = self.data_scat_coef_vectorized.copy()
        if exclude_timestamps is not None:
            for ts in exclude_timestamps:
                print(
                    f'Timestamp {UTCDateTime(ts)} - {UTCDateTime(ts) + (self.network_segment * exclude_timestamps_skip)} has been excluded from the Scaling fit and Model fit.'
                )
                ts_index = self._get_index_from_UTC_timestamp(ts)
                scat_coeff_data = np.delete(scat_coeff_data, np.s_[ts_index:ts_index + exclude_timestamps_skip], 0)

        scaler = RobustScaler()
        # fit on subset
        scaler.fit(scat_coeff_data)
        # Transform on full data
        scat_coeff_data_scaled = scaler.transform(self.data_scat_coef_vectorized.copy())

        print('Fitting and transforming the data for FastICA')
        # fit on a subset of data
        model.fit(scaler.transform(scat_coeff_data))
        # Transform on full data
        features = model.transform(scat_coeff_data_scaled)

        features_inverse = model.inverse_transform(features)
        score_exp_var = explained_variance_score(scat_coeff_data_scaled, features_inverse)
        score_mse = mean_squared_error(scat_coeff_data_scaled, features_inverse)
        print(f'      ICAs #{num_ICA} : Explained Variance {score_exp_var * 100:0.5f}% : MSE {score_mse:0.5f}')

        self.ica = model
        self.ica_number_components = num_ICA
        self.ica_features = features
        self.ica_explained_variance_score = score_exp_var

        # SAVE ICA MODEL IN PICKLE FILE
        with open(ica_model_path, 'wb') as handle:
            pickle.dump(model, handle, protocol=pickle.HIGHEST_PROTOCOL)

        # save the features
        np.savez(
            f'{self.data_savepath}data/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_features_{num_ICA}.npz',
            features=features,
            features_inverse=features_inverse,
            score_explained_variance=score_exp_var,
            score_mse=score_mse)

        if return_data:
            return score_exp_var, score_mse, model, features

    def process_ICA_range(self,
                          exclude_timestamps: Optional[list[str]] = None,
                          exclude_timestamps_skip: int = 3,
                          **kwargs) -> None:
        """Process ICA dimension reduction for a range of possible dimensions controlled via ica_min_ICAs, ica_max_ICAs and ica_ev_limit.
        """
        score_exp_var = 0
        ica_min_ICAs = 6 if self.ica_min_ICAs is None else self.ica_min_ICAs
        ica_max_ICAs = 20 if self.ica_max_ICAs is None else self.ica_max_ICAs
        num_ICA = ica_min_ICAs - 1
        while (score_exp_var <= self.ica_ev_limit and num_ICA < ica_max_ICAs):
            num_ICA += 1
            score_exp_var, _, model, features = self.process_ICA_single(num_ICA=num_ICA,
                                                                        return_data=True,
                                                                        exclude_timestamps=exclude_timestamps,
                                                                        exclude_timestamps_skip=exclude_timestamps_skip,
                                                                        **kwargs)
        if num_ICA == ica_max_ICAs:
            print(
                f'Failed to reach an explained variance of {self.ica_ev_limit* 100:0.5f} with a max number of {self.ica_max_ICAs} ICAs. Either increase the maximum number of ICAs or decrease the Explained Variance Limit.'
            )
        else:
            print(f'Recommended number of ICAs {num_ICA}.\nThe Explained Variance with these ICAs is {score_exp_var}.')

        self.ica = model
        self.ica_number_components = num_ICA
        self.ica_features = features
        print(f'Compressed Vectorised Scat. Coefficients Array after ICA : {features.shape}')

    def preload_ICA(self, num_ICA: int) -> None:
        """Load a pre-calculated ICA and set required variables

        Args:
            num_ICA (int): Desired number of ICAs
        """
        if not os.path.exists(
                f'{self.data_savepath}ICA/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_dimension_{num_ICA}.pickle'
        ):
            raise ValueError(
                f"The supplied number of ICAs {num_ICA} has not been computed. Choose another number for the ICAs or calcate using 'process_ICA_single'"
            )

        self.ica = pickle.load(
            open(
                f'{self.data_savepath}ICA/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_dimension_{num_ICA}.pickle',
                'rb'))

        ica_results = np.load(
            f'{self.data_savepath}data/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_features_{num_ICA}.npz'
        )
        self.ica_features = ica_results['features']
        self.ica_features_inverse = ica_results['features_inverse']
        self.ica_explained_variance_score = ica_results['score_explained_variance']
        self.ica_mse = ica_results['score_mse']

        self.ica_number_components = num_ICA

        print(f'Compressed Vectorised Scat. Coefficients Array after ICA : {self.ica_features.shape}')

        self.data_times = np.load(
            f'{self.data_savepath}data/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_times.npy'
        )

    def plot_ICA(self, **kwargs) -> None:
        """Visualise the ICAs
        """
        dimensions = self.ica_features.shape[1]
        # Preprocess
        features = self.ica_features.T
        features = demad(features)

        # features = latents
        FACTOR = 0.5

        # Show in time
        n_features, _ = features.shape

        # Figure
        kwargs['figsize'] = (10, 7) if kwargs.get('figsize') is None else kwargs.get('figsize')
        _, ax = plt.subplots(1, **kwargs)

        # Show
        for index, feature in enumerate(features.copy()):
            color = f'C{index % 3}'
            feature *= 0.5
            feature += index + 1
            feature_filtered = medfilt(
                feature, self.ica_median_filter if self.ica_median_filter is not None else
                (int(3600 / self.network_segment) * 3) + 1)
            ax.plot(self.data_times, feature, '.', ms=3, alpha=0.5, mew=0, color=color)
            ax.plot(self.data_times, feature_filtered, lw=0.7, color='k')

        # Labels
        ax.grid()
        ax.set_ylim(0, n_features + FACTOR)
        ax.set_yticks(np.arange(n_features) + 1)
        ax.set_yticklabels(list_of_strings(n_features))
        ax.set_ylabel('Independant component index')

        # Date labels
        dateticks = mdates.AutoDateLocator()
        datelabels = mdates.ConciseDateFormatter(dateticks)
        ax.xaxis.set_major_locator(dateticks)
        ax.xaxis.set_major_formatter(datelabels)
        ax.set_xlim(self.data_times.min(), self.data_times.max())

        # Remove borders
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.tick_params(axis='y', length=0)
        plt.title(
            f'{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_ICAs_{dimensions}\n{dimensions} ICAs - Explained Variance : {self.ica_explained_variance_score*100:0.5f}'
        )

        plt.savefig(
            f'{self.data_savepath}figures/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_ICAs_{dimensions}.png'
        )

        plt.show()

    def plot_ICA_zoom(self, ICA_letter: str, zoom_time_start: str, zoom_time_end: str, **kwargs) -> None:
        """Visualise a zoomed section of an ICA

        Args:
            ICA_letter (str): ICA letter
            zoom_time_start (str): Zoom section start
            zoom_time_end (str): Zoom section end
        """
        ica_number = [
            x_enum for x_enum, x in enumerate(list_of_strings(self.ica.n_components)) if x == ICA_letter.upper()
        ][0]

        features = self.ica_features.T
        features_smoothed = demad(features)
        # Zoom into an ICA
        kwargs['figsize'] = (10, 7) if kwargs.get('figsize') is None else kwargs.get('figsize')
        _, ax = plt.subplots(**kwargs)
        ax.plot(self.data_times, features[ica_number], '.', ms=3, alpha=0.5, mew=0)
        ax.plot(self.data_times, features_smoothed[ica_number], lw=0.7, color='k')
        ax.set_xlim(min(self.data_times[self.data_times > int(mdates.date2num(UTCDateTime(zoom_time_start)))]),
                    min(self.data_times[self.data_times > int(mdates.date2num(UTCDateTime(zoom_time_end)))]))
        ax.set_ylim(-2, 2)
        ax.grid()
        locator = mdates.AutoDateLocator()
        formatter = mdates.ConciseDateFormatter(locator)
        ax.xaxis.set_major_locator(locator)
        ax.xaxis.set_major_formatter(formatter)

        plt.savefig(
            f'{self.data_savepath}figures/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_ICAs_{self.ica.n_components}_{ICA_letter}_Zoom_{zoom_time_start}_{zoom_time_end}.png'
        )

        plt.show()

    def plot_ica_contribution(self, **kwargs) -> None:
        """Visualise the ICA contribution to each cluster
        """

        # Calculate centroid
        classes = np.unique(self.dendrogram_predictions)
        feature_usage = np.zeros((self.ica_features.shape[1], len(classes)))

        for cluster in classes:
            within_cluster = self.dendrogram_predictions == cluster
            cluster_samples = self.ica_features[within_cluster]
            contributions = np.mean(cluster_samples, axis=0)
            contributions /= np.abs(contributions).sum()
            feature_usage[:, cluster - 1] = contributions

        vmax = round_nearest(np.abs(feature_usage).max(), 0.05)
        cmap = plt.cm.RdYlBu
        bounds = np.arange(-vmax, vmax, .05)
        norm = mpl.colors.BoundaryNorm(bounds, cmap.N)

        kwargs['figsize'] = (8, 5) if kwargs.get('figsize') is None else kwargs.get('figsize')
        _, ax = plt.subplots(1, 1, **kwargs)
        img = ax.matshow(feature_usage, cmap=cmap, norm=norm)

        cb = plt.colorbar(img, orientation='horizontal')
        cb.minorticks_on()
        cb.set_label('Normalised contribution')
        ax.set_title('Contribution of ICA to prediction\nCluster')
        ax.set_ylabel('ICA component')
        ax.set_xticks(range(len(classes)), range(1, len(classes) + 1))
        ax.set_yticks(range(self.ica_features.shape[1]), list_of_strings(self.ica_features.shape[1]))

        plt.savefig(
            f'{self.data_savepath}figures/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_ICA_{self.ica.n_components}_clustering_{self.ica.n_components}_Waveform_contribution.png',
            bbox_inches='tight')

        plt.show()

    def list_linkages(self):
        print(
            glob(
                f'{self.data_savepath}clustering/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_ICA_*linkage*'
            ))



import datetime
import pickle

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from obspy import UTCDateTime
from obspy.signal.cross_correlation import correlate, xcorr_max
from tqdm import tqdm

# EXAMPLE
# calculate_waveform_correlations(sort_type ='distance_filter', sort_filter=100, envelope=True)
# plot_correlation_waveforms(correlations=correlations_waveforms_distance_100, sort_type ='distance_filter',  sort_filter=100, envelope=False)


class WaveformCorrelations:

    def process_cluster_trace_correlation(self,
                                          df_preds: pd.DataFrame,
                                          cluster: int = 1,
                                          sort_type: str = 'all',
                                          sort_filter: int = None,
                                          time_second: int = 60,
                                          channel: str = 'HHZ',
                                          envelope: bool = False):
        # sort_type = ['all', 'xcorr_filter', 'distance_filter']

        df_times = df_preds.loc[df_preds.predictions == cluster, ['cluster_rank', 'times_YYYYMMDD']]
        time_list = list(df_times.sort_values(by='cluster_rank')['times_YYYYMMDD'].values)

        if sort_type == 'distance_filter':
            if sort_filter is None:
                msg = 'If using \'sort_type\' => distance_filter, you need to supply \'sort_filter\' of type integer. \n'+\
                        'This will be used to take the waveforms up to the n\'th based on the Euclidean distance \n'+\
                        'from the cluster centroid'
                raise ValueError(msg)
            time_list = time_list[:sort_filter + 1]

        wvf0_time = time_list.pop(0)

        wvf0 = self.get_waveform(start_time=str(wvf0_time), time_second=time_second, channel=channel, envelope=envelope)
        if len(wvf0) < 1:
            print(f'  Waveform cross-correlations cannot be computed for Cluster {cluster}.\n' +
                  '  The waveform in the centre of the cluster contains no traces. This \n' +
                  '  is most likely a cluster of damaged/missing data.')
            # return None
        else:
            time_per_sample = 1 / wvf0[0].stats.sampling_rate
            wvf0 = wvf0.select(component='Z')[0].data

            corr = []
            for st_enum, st in tqdm(enumerate(time_list)):
                wvf1 = self.get_waveform(start_time=str(st),
                                         time_second=time_second,
                                         channel=channel,
                                         envelope=envelope)
                if len(wvf1) > 0:
                    wvf1 = wvf1.select(component='Z')[0].data
                    cc = correlate(wvf0, wvf1, len(wvf0))
                    shift, value = xcorr_max(cc)
                    _sort_filter = 0
                    if sort_type == 'xcorr_filter':
                        if sort_filter > 1:
                            msg = 'If using \'sort_type\' => xcorr_filter, you need to \'supply sort_filter\' \n'+\
                                'of type float corresponding to the absolute maximum cross-correlation \n'+\
                                'coefficient to include in analysis.'
                            raise ValueError(msg)
                        _sort_filter = sort_filter
                    if abs(value) >= _sort_filter:
                        time_start2 = UTCDateTime(str(st)) + datetime.timedelta(seconds=-1 * shift * time_per_sample)
                        shifted_correlated_waveform = [-1 if value < 0 else 1] * self.get_waveform(
                            start_time=time_start2, envelope=envelope)[0].data
                        corr.append({
                            'cluster_rank': st_enum + 1,
                            'correlation_value': value,
                            'correlation_shift': shift,
                            'waveform_time': str(st),
                            'waveform': wvf1,
                            'shifted_correlated_waveform_time': str(time_start2),
                            'shifted_correlated_waveform': shifted_correlated_waveform
                        })

            return {'centre_waveform_time': str(wvf0_time), 'centre_waveform': wvf0, 'correlations': corr}

    def calculate_waveform_correlations(self,
                                        df_preds: pd.DataFrame,
                                        sort_type='distance_filter',
                                        sort_filter=100,
                                        time_second=60,
                                        channel='HHZ',
                                        envelope=False):

        correlations = {}
        for cluster in set(df_preds.predictions):
            corr = self.process_cluster_trace_correlation(df_preds, cluster, sort_type, sort_filter, time_second,
                                                          channel, envelope)
            correlations[cluster] = corr

        _wvf_type = 'waveform'
        if envelope:
            _wvf_type = 'envelope'
        save_file = f'{self.data_savepath}data/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_ICA_{self.ica.n_components}_clustering_{self.ica.n_components}_{_wvf_type}_correlations_{sort_type}_{sort_filter}.pkl'
        with open(save_file, 'wb') as handle:
            pickle.dump(correlations, handle, protocol=pickle.HIGHEST_PROTOCOL)

        return correlations

    def load_correlations(self, sort_type='distance_filter', sort_filter=100, envelope=False):
        _wvf_type = 'waveform'
        if envelope:
            _wvf_type = 'envelope'
        save_file = f'{self.data_savepath}data/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_ICA_{self.ica.n_components}_clustering_{self.ica.n_components}_{_wvf_type}_correlations_{sort_type}_{sort_filter}.pkl'

        with open(save_file, 'rb') as handle:
            correlations = pickle.load(handle)

        return correlations

    def stack_correlations(self, correlations, cluster):
        if correlations[cluster] is None:
            return None
        else:
            waveforms = [x['shifted_correlated_waveform'] for x in correlations[cluster]['correlations']]
            return np.mean(waveforms, axis=0)

    def process_waveform_correlations_stacked_waveform(self,
                                                       df_preds,
                                                       correlations,
                                                       sort_type,
                                                       sort_filter,
                                                       envelope=False):
        correlation_waveform = {}
        for cluster in set(df_preds.predictions):
            correlation_waveform[cluster] = self.stack_correlations(correlations, cluster)

        _wvf_type = 'waveform'
        if envelope:
            _wvf_type = 'envelope'
        np.save(
            f'{self.data_savepath}data/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_ICA_{self.ica.n_components}_clustering_{self.ica.n_components}_{_wvf_type}_correlations_stacked_waveform_{sort_type}_{sort_filter}.npy',
            correlation_waveform)

        return correlation_waveform

    def plot_correlation_waveforms(self, df_preds, correlations, sort_type, sort_filter, envelope=False):

        correlation_waveform = self.process_waveform_correlations_stacked_waveform(df_preds, correlations, sort_type,
                                                                                   sort_filter, envelope)

        fig, ax = plt.subplots(len(set(df_preds.predictions)), 1, figsize=(20, 10), sharex=True, sharey=True)
        for clust_enum, cluster in enumerate(correlations.keys()):
            if correlations[cluster] is None:
                ax[clust_enum].set_title(f'Cluster {clust_enum+1} - Empty Traces')
            else:
                lines = []
                num_traces = len(correlations[cluster]['correlations'])
                avg_xcorr = np.mean([x['correlation_value'] for x in correlations[cluster]['correlations']])
                ax[clust_enum].set_title(
                    f'Cluster {clust_enum+1} : Number of Traces {num_traces} : Avg. Cross-correlation Coeff. {avg_xcorr:.3}'
                )
                for x in correlations[cluster]['correlations'][:100]:
                    lines += ax[clust_enum].plot(x['shifted_correlated_waveform'],
                                                 color='b',
                                                 alpha=0.2,
                                                 linewidth=1,
                                                 label='shifted_correlated_waveforms')

                centre_waveform = correlations[cluster]['centre_waveform']
                centre_waveform_shift = np.mean(np.abs(centre_waveform)) * .5
                lines += ax[clust_enum].plot(centre_waveform,
                                             color='k',
                                             alpha=0.3,
                                             linewidth=1,
                                             label='centre_waveform')
                ax[clust_enum].set_ylim(
                    [min(centre_waveform) - centre_waveform_shift,
                     max(centre_waveform) + centre_waveform_shift])

                lines += ax[clust_enum].plot(correlation_waveform[cluster],
                                             color='r',
                                             linewidth=0.5,
                                             label='correlation_waveform')

        _wvf_type = 'Waveform'
        if envelope:
            _wvf_type = 'Envelope'

        fig.legend(lines[-3:],
                   [f'Shifted and corrected {_wvf_type}s', f'Centroid {_wvf_type}', f'Correlation {_wvf_type}'],
                   loc='upper center',
                   ncol=3,
                   bbox_to_anchor=(0.5, -0.01))
        ax[0].margins(x=0)

        file_name = f'{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_ICA_{self.ica.n_components}_clustering_{self.ica.n_components}_{_wvf_type}_Correlations_{sort_type}_{sort_filter}'
        plt.suptitle(file_name)
        plt.savefig(f'{self.data_savepath}figures/{file_name}.png', bbox_inches='tight')
        plt.show()

    def plot_correlation_frequency(self, df_preds, correlations_waveforms_distance_all):
        fig, ax = plt.subplots(len(set(df_preds.predictions)), 1, figsize=(20, 10), sharex=True, sharey=True)
        for clust_enum, cluster in enumerate(correlations_waveforms_distance_all.keys()):
            if correlations_waveforms_distance_all[cluster] is None:
                ax[clust_enum].set_title(f'Cluster {clust_enum+1} - Empty Traces')
            else:
                data = [
                    np.abs(xcor['correlation_value'])
                    for xcor in correlations_waveforms_distance_all[cluster]['correlations']
                ]
                sns.histplot(data, ax=ax[clust_enum])
                ax[clust_enum].set_title(f'Cluster {clust_enum+1} ')

        plt.show()

    def plot_correlation_shift(self, correlations: dict, clusters: list, within_cluster_number: int = None):

        _within_cluster_number = 100 if within_cluster_number is None else within_cluster_number

        fig, ax = plt.subplots(2, len(clusters), figsize=(20, 10), sharex=True, sharey=True)
        for cluster_enum, cluster in enumerate(clusters):
            ax[0, cluster_enum].plot(correlations[cluster]['centre_waveform'], 'k')
            for x_enum, x in enumerate(correlations[cluster]['correlations'][:_within_cluster_number]):
                ax[1, cluster_enum].plot(x['waveform'] + ((x_enum + 1) * 0.0000001), 'b', alpha=0.2)
            plt.title(f'Cluster {cluster}')

        file_name = f'{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_ICA_{self.ica.n_components}_clustering_{self.ica.n_components}_{_wvf_type}_Correlations'
        plt.suptitle(f'{file_name}\nWaveform Correlations')
        plt.savefig(f'{self.data_savepath}figures/{file_name}_Cluster_waveform_correlations.png', bbox_inches='tight')
        plt.show()



import os

import matplotlib.pyplot as plt
import numpy as np
import obspy
import obspy.signal.filter
from matplotlib import dates as mdates
from obspy import Stream
from tqdm import tqdm


class AnalysisProcessing():

    def process_stream_envelope(self, stream: Stream) -> Stream:
        st_env = stream.copy()
        for trace_enum, tr in enumerate(st_env):
            st_env[trace_enum].data = obspy.signal.filter.envelope(tr.data)

        return st_env

    def process_stream_rms(self, stream: Stream) -> Stream:
        return [np.sqrt(np.mean(tr.data**2)) for tr in stream]

    def build_waveforms_data(self, include_envelopes=False, include_RMS=False, custom_title=''):
        waveforms_file = f'{self.data_savepath}data/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_segment}_waveforms{custom_title}.npy'
        if os.path.exists(waveforms_file):
            print('Waveforms already calculated. Loading into self.waveforms')
            waveforms = self.load_waveforms(custom_title)
        else:
            self.load_data_times()
            start_times = [obspy.UTCDateTime(s_time) for s_time in mdates.num2date(self.data_times)]
            end_times = [e_time + self.network_segment for e_time in start_times]

            data_stream = []
            data_envelope = []
            data_rms = []

            for start, end in tqdm(zip(start_times, end_times)):
                ds = self.load_data(starttime=start, endtime=end, channel=self.data_channel)
                data_stream.append(ds)
                if include_envelopes:
                    data_envelope.append(self.process_stream_envelope(ds))
                if include_RMS:
                    data_rms.append(self.process_stream_rms(ds))
            waveforms = np.array(data_stream)
            np.save(waveforms_file, waveforms)

            if include_envelopes:
                np.save(
                    f'{self.data_savepath}data/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_segment}_envelopes{custom_title}.npy',
                    np.array(data_envelope))

            if include_RMS:
                np.save(
                    f'{self.data_savepath}data/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_segment}_RMS{custom_title}.npy',
                    np.array(data_rms))

        self.waveforms = waveforms
        return waveforms

    def load_waveforms(self, waveform_extra_title: str = ''):
        waveforms = np.load(
            f'{self.data_savepath}data/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_segment}_waveforms{waveform_extra_title}.npy',
            allow_pickle=True)
        self.waveforms = waveforms
        return waveforms

    def get_spectrogram_data(self, Zwaveform, samp_rate):
        if len(Zwaveform.traces) > 0:
            spec = Zwaveform.spectrogram(samp_rate=samp_rate, show=False)
            orig_data = spec[0].axes[0].images[0].get_array().data
            plt.close()
            return orig_data
        else:
            return ''

    def spectrogram_build(self, waveforms, samp_rate, extra_title=''):
        spectrum_file = f'{self.data_savepath}data/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_segment}_spectrograms{extra_title}.npy'
        if os.path.exists(spectrum_file):
            print(f'>> Spectrum already exists for {spectrum_file}')
            spectrograms = self.load_spectrograms(extra_title)
        else:
            specgrams = []
            for waveform_enum, waveform in tqdm(enumerate(waveforms)):
                specgrams.append(self.get_spectrogram_data(waveform.select(component='Z'), samp_rate))
            spectrograms = np.array(specgrams)
            np.save(spectrum_file, spectrograms)

        self.spectrograms = spectrograms

        return spectrograms

    def build_spectrum_data(
        self,
        waveform_extra_title: str = '',
        spectrum_extra_title: str = '',
        spectrum_samp_rate: int = 50,
    ):
        if self.waveforms is None:
            self.waveforms = self.load_waveforms(waveform_extra_title)
        self.spectrogram_build(waveforms=self.waveforms, samp_rate=spectrum_samp_rate, extra_title=spectrum_extra_title)

    def load_spectrograms(self, extra_title=''):
        spectrograms = np.load(
            f'{self.data_savepath}data/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_segment}_spectrograms{extra_title}.npy',
            allow_pickle=True)
        self.spectrograms = spectrograms

        return spectrograms



import datetime
from glob import glob
from typing import List

import matplotlib.pyplot as plt
import numpy as np
import obspy
import pandas as pd
from matplotlib import dates as mdates
from obspy.core import UTCDateTime
from scipy.spatial.distance import euclidean

from scatcluster.helper import COLORS, is_gpu


class Waveforms:

    def plot_waveforms_per_cluster(self, clusters: List[int] = None, waveforms_n_samples: int = None, **kwargs):
        """Visualise Waveforms for a All or a select number of clusters.

        Args:
            clusters (List[int], optional): If specified, only plot waveforms for clusters in list of integers.
            For example, clusters = [1,3] will produce waveforms for cluster 1 and 3 only.
            Defaults to None will produce waveforms for all clusters.
        """
        # get channel list from plotting
        stream = self.load_data(starttime=UTCDateTime(self.data_sample_starttime),
                                endtime=UTCDateTime(self.data_sample_endtime),
                                channel=self.data_channel)
        channel_list = sorted(list(set([trace.stats.channel for trace in stream])))

        if clusters is not None:
            classes = clusters
        else:
            # Calculate centroid
            classes = np.unique(self.dendrogram_predictions)

        for cluster in classes:

            # Extract clusters
            within_cluster = self.dendrogram_predictions == cluster
            cluster_samples = self.ica_features[within_cluster]
            cluster_times = self.dendrogram_timestamps[within_cluster]

            # Centroid
            centroid = np.median(cluster_samples, axis=0)
            distances = list()
            for sample in cluster_samples:
                distances.append(euclidean(sample, centroid))
            distances = np.array(distances)

            # Extract best waveforms timestamps
            _waveforms_n_samples = 5 if waveforms_n_samples is None else waveforms_n_samples
            distances_argsort = np.argsort(distances)
            sorted_times = cluster_times[distances_argsort][:_waveforms_n_samples]
            sorted_times = mdates.num2date(sorted_times)
            kwargs['figsize'] = (20, 8) if kwargs.get('figsize') is None else kwargs.get('figsize')
            _, ax = plt.subplots(1, 4, gridspec_kw={'width_ratios': [2, 2, 2, 1]}, **kwargs)
            for ch, channel in enumerate(channel_list):
                yticks = list()
                for index, t in enumerate(sorted_times):
                    start = obspy.UTCDateTime(t)
                    end = start + self.network_segment
                    stream = self.load_data(starttime=start, endtime=end, channel=self.data_channel)
                    if len(stream) > 0:
                        trace = stream[0]
                        data = trace.data
                        data /= np.abs(data).max()
                        ax[ch].plot(trace.times(), data + index, lw=0.3)
                        yticks.append(start)
                    else:
                        print(f'Trace is not available between {start}-{end} for supplied parametrization')
                ax[ch].set_title(f'channel {channel}')
            ax[0].set_yticks(range(len(yticks)), [x.strftime('%Y-%m-%d %H:%M:%S') for x in yticks])
            self.show_dendrogram_waveforms(linkage=self.dendrogram_linkage,
                                           ax=ax[3],
                                           depth=len(np.unique(self.dendrogram_predictions)))
            ax[3].set_title('Dendrogram')
            self._set_share_axes(axs=ax[:2], sharex=True, sharey=True)
            plt.suptitle(
                f'{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_ICA_{self.ica.n_components}_clustering_{clusters}\nCluster {cluster}'
            )

            plt.savefig(
                f'{self.data_savepath}figures/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_ICA_{self.ica.n_components}_clustering_{clusters}_waveforms_cluster_{cluster}.png'
            )

            plt.show()

    def plot_all_waveforms(self, channel='HHZ', waveforms_n_samples=3, **kwargs):
        classes = np.unique(self.dendrogram_predictions)
        _waveforms_n_samples = 5 if waveforms_n_samples is None else waveforms_n_samples
        kwargs['figsize'] = (20, 8) if kwargs.get('figsize') is None else kwargs.get('figsize')
        _, ax = plt.subplots(
            _waveforms_n_samples,
            len(classes),
            # sharey=True,
            sharex=True,
            #  gridspec_kw={'width_ratios': [2, 2, 2, 1]},
            **kwargs)

        for cluster_enum, cluster in enumerate(classes):
            # Extract clusters
            within_cluster = self.dendrogram_predictions == cluster
            cluster_samples = self.ica_features[within_cluster]
            cluster_times = self.dendrogram_timestamps[within_cluster]

            # Centroid
            centroid = np.median(cluster_samples, axis=0)
            distances = list()
            for sample in cluster_samples:
                distances.append(euclidean(sample, centroid))
            distances = np.array(distances)

            # Extract best waveforms timestamps

            distances_argsort = np.argsort(distances)
            sorted_times = cluster_times[distances_argsort][:_waveforms_n_samples]
            sorted_times = mdates.num2date(sorted_times)

            for index, t in enumerate(sorted_times):
                start = obspy.UTCDateTime(t)
                end = start + self.network_segment
                stream = self.load_data(starttime=start, endtime=end, channel=channel)
                if len(stream) > 0:
                    trace = stream[0]
                    stream_data = trace.data
                    ax[index, cluster_enum].plot(trace.times(), stream_data, lw=0.3, color=COLORS[cluster_enum])
                    ax[index, cluster_enum].annotate(start.strftime('%Y/%m/%d %H:%M:%S'),
                                                     xy=(0, 1),
                                                     xycoords='axes fraction',
                                                     horizontalalignment='left',
                                                     verticalalignment='top')
                else:
                    print(f'Trace is not available between {start}-{end} for supplied parametrization')

            ax[0, cluster_enum].set_title(f'Cluster {cluster}')
        [axc.set_axis_off() for axr in ax for axc in axr]
        plt.subplots_adjust(wspace=0, hspace=0)
        plt.suptitle(
            f'{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_ICA_{self.ica.n_components}_clustering_{len(classes)}'
        )

        plt.savefig(
            f'{self.data_savepath}figures/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_ICA_{self.ica.n_components}_clustering_{len(classes)}_all_clusters_waveforms.png'
        )

        plt.show()

    def get_waveform(self, start_time='2022-01-01T00:00:00+00:00', time_second=60, channel='HHZ', envelope=False):
        time_start = UTCDateTime(str(start_time))
        time_end = UTCDateTime(str(start_time)) + datetime.timedelta(seconds=time_second)
        wvf = self.load_data(starttime=time_start, endtime=time_end, channel=channel)
        if envelope:
            wvf = self.process_stream_envelope(wvf)
        return wvf

    def retreive_scat_vector(self,
                             scat_coef_vectorized,
                             scat_coef_vectorized_index,
                             num_channels=3,
                             scat_vec_size_first_order=16,
                             scat_vec_size_second_order=(16, 7)):
        scat_vec = scat_coef_vectorized[scat_coef_vectorized_index, :]
        scat_vec_first_order = scat_vec[:(num_channels * scat_vec_size_first_order)].reshape(
            num_channels, scat_vec_size_first_order)
        scat_vec_second_order = scat_vec[(num_channels * scat_vec_size_first_order):].reshape(
            num_channels, scat_vec_size_second_order[0], scat_vec_size_second_order[1])
        return scat_vec_first_order, scat_vec_second_order

    def process_waveforms_envelopes_scat_coefficients(self, df_preds):
        self.load_scat_coef_vectorized()
        file_list = glob(
            f'{self.data_savepath}scatterings/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_scatterings_*.npz'
        )
        scat_file = np.load(file_list[0])
        print(f"First order : {scat_file['scat_coef_0'].shape[1:]}")
        print(f"Second order : {scat_file['scat_coef_1'].shape[1:]}")

        num_channels = scat_file['scat_coef_0'].shape[1:][0]
        first_order = scat_file['scat_coef_0'].shape[1:][1]
        second_order = scat_file['scat_coef_1'].shape[1:][1:]

        scat_vec_list = [
            self.retreive_scat_vector(self.scat_coef_vectorized, i, num_channels, first_order, second_order)
            for i in df_preds.index
        ]

        df_scat_vec = pd.DataFrame({
            'scat_vec_first_order': [sv[0] for sv in scat_vec_list],
            'scat_vec_second_order': [sv[1] for sv in scat_vec_list]
        })
        return df_scat_vec

    def waveforms_envelopes_scat_coefficients_df_merge(
        self,
        df_preds,
        df_scat_vec,
    ):
        return pd.concat([df_preds, df_scat_vec], axis=1)

    def plot_waveforms_envelopes_scat_coefficients(self,
                                                   df_pred_scat_vec: pd.DataFrame,
                                                   channel: str = 'HHZ',
                                                   num_waveforms: int = 5,
                                                   num_scat_coeff_stacking: int = 10,
                                                   GPU: bool = is_gpu()):
        if GPU:
            self.net.banks[0].centers = self.net.banks[0].centers.get()
            self.net.banks[1].centers = self.net.banks[2].centers.get()

        fig, ax = plt.subplots(len(set(df_pred_scat_vec.predictions)), 4, figsize=(20, 10), sharex='col', sharey='col')

        sv1_mean = np.mean(np.array([sv1[2, :] for sv1 in df_pred_scat_vec.scat_vec_first_order]), axis=0)
        sv2_mean = np.mean(np.array([sv2[2, :, :] for sv2 in df_pred_scat_vec.scat_vec_second_order]), axis=0)

        ax[0, 0].set_title('Waveforms')
        ax[0, 0].margins(x=0)
        ax[len(set(df_pred_scat_vec.predictions)) - 1, 0].set_xlabel(f'Time samples ({self.network_segment}s)')

        ax[0, 1].set_title('Envelopes')
        ax[0, 1].margins(x=0)
        ax[len(set(df_pred_scat_vec.predictions)) - 1, 1].set_xlabel(f'Time samples ({self.network_segment}s)')

        ax[0, 2].set_title('First Order Scat. Coeff.')
        # ax[0,2].set_xscale('log')
        # ax[0,2].set_yscale('log')
        ax[len(set(df_pred_scat_vec.predictions)) - 1, 2].set_xlabel('First-order frequency (Hz)')
        ax[len(set(df_pred_scat_vec.predictions)) - 1, 2].set_xticks(range(0, sv1_mean.shape[0]))
        first_order_scat = [f'{x:.1f}' for x in reversed(self.net.banks[0].centers)]
        ax[len(set(df_pred_scat_vec.predictions)) - 1, 2].set_xticklabels(labels=first_order_scat, rotation='vertical')

        ax[0, 3].set_title('Second Order Scat. Coeff.')
        # ax[0,3].set_xscale('log')
        # ax[0,3].set_yscale('log')
        ax[len(set(df_pred_scat_vec.predictions)) - 1, 3].set_xlabel('First-order frequency (Hz)')
        ax[len(set(df_pred_scat_vec.predictions)) - 1, 3].set_xticks(range(0, sv1_mean.shape[0]))
        ax[len(set(df_pred_scat_vec.predictions)) - 1, 3].set_xticklabels(labels=first_order_scat, rotation='vertical')
        ax[len(set(df_pred_scat_vec.predictions)) - 1, 3].set_yticks(range(0, sv2_mean.shape[1]))
        second_order_scat = [f'{x:.1f}' for x in self.net.banks[1].centers]
        ax[len(set(df_pred_scat_vec.predictions)) - 1, 3].set_yticklabels(labels=second_order_scat)

        for row in range(len(set(df_pred_scat_vec.predictions))):
            ax[row, 2].set_ylabel('First-order scat. coef.')
            ax[row, 3].set_ylabel('Second-order freq. (Hz)')

        num_clusters = list(set(df_pred_scat_vec.predictions))
        for cluster_enum, cluster in enumerate(num_clusters):

            ax[cluster_enum, 0].set_ylabel(f'Cluster {cluster}')

            wvf0_time = str(df_pred_scat_vec.loc[(df_pred_scat_vec.predictions == cluster) &
                                                 (df_pred_scat_vec.cluster_rank == 0), 'times_YYYYMMDD'].values[0])
            wvf0 = self.get_waveform(start_time=str(wvf0_time),
                                     time_second=self.network_segment,
                                     channel=channel,
                                     envelope=False)
            if len(wvf0) < 1:
                print(f'  The waveform in the centre of Cluster {cluster} contains no traces. This \n' +
                      '  is most likely a cluster of damaged/missing data.')
            else:
                df_scat_vec = df_pred_scat_vec.loc[
                    (df_pred_scat_vec.predictions == cluster),
                ].sort_values(by='cluster_rank').reset_index()

                # WAVEFORMS

                ax[cluster_enum, 0].plot(self.min_max_vector(wvf0[0].data) - 0.5, 'k', alpha=0.5)
                wvfs = [
                    self.get_waveform(start_time=str(st).replace(' ', 'T'),
                                      time_second=self.network_segment,
                                      channel=channel,
                                      envelope=False)[0].data for st in df_scat_vec['times_YYYYMMDD'][1:num_waveforms]
                ]
                for wvf_enum, wvf in enumerate(wvfs):
                    ax[cluster_enum, 0].plot(self.min_max_vector(wvf) + wvf_enum + 1 - 0.5, 'b', alpha=0.2)

                # ENVELOPES
                env0 = self.get_waveform(start_time=str(wvf0_time),
                                         time_second=self.network_segment,
                                         channel=channel,
                                         envelope=True)
                ax[cluster_enum, 1].plot(self.min_max_vector(env0[0].data), 'k', alpha=0.5)
                envs = [
                    self.get_waveform(start_time=str(st).replace(' ', 'T'),
                                      time_second=self.network_segment,
                                      channel=channel,
                                      envelope=True)[0].data for st in df_scat_vec['times_YYYYMMDD'][1:num_waveforms]
                ]
                for env_enum, env in enumerate(envs):
                    ax[cluster_enum, 1].plot(self.min_max_vector(env) + env_enum + 1, 'b', alpha=0.2)

                # First Order Scat Coeffs
                sv1_array = np.array([sv1[2, :] for sv1 in df_scat_vec.scat_vec_first_order[:num_scat_coeff_stacking]])
                sv1_array = sv1_array - sv1_mean
                for sv1_row in range(sv1_array.shape[0]):
                    ax[cluster_enum, 2].plot(sv1_array[sv1_row, :], 'b', alpha=0.02)
                ax[cluster_enum, 2].plot(np.mean(sv1_array, axis=0), 'k', alpha=0.5)

                # Second Order Scat Coeffs
                sv2_array = np.array(
                    [sv2[2, :, :] for sv2 in df_scat_vec.scat_vec_second_order[:num_scat_coeff_stacking]])
                sv2_array = sv2_array - sv2_mean
                sv2_array = np.mean(sv2_array, axis=0).T
                sv2_array = np.flip(sv2_array, 0)
                print(f'MIN {sv2_array.min()} - MAX {sv2_array.max()}')
                im = ax[cluster_enum, 3].imshow(sv2_array, aspect='auto', cmap='RdYlBu')
                plt.colorbar(im, ax=ax[cluster_enum, 3])

        filename = f'{self.data_network}_{self.data_station}_{self.data_location}_{self.network_name}_ICA_{self.ica.n_components}_Clustering_{len(num_clusters)}_Xcorr_ScatCoeff'
        plt.suptitle(filename)
        plt.savefig(f'{self.data_savepath}figures/{filename}.png', bbox_inches='tight')
        plt.show()



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.axes import Axes
from scipy.cluster import hierarchy
from sklearn.metrics import fowlkes_mallows_score

from scatcluster.helper import COLORS


class SSNCrossTabAnalysis:
    """ScatCluster Add-on to run cross-analysis of different clusterings
    """

    def __init__(
        self,
        data_savepath: str = '/home/jovyan/shared/users/zerafa/data/sds.chris/scatcluster-sds/',
        data_network: str = 'ET',
        data_station: str = 'SOE0',
        data_location: str = '',
        network_sampling_rate_banks_pooling: str = '50_4_4_2_7_1_1_avg',
        ica_number: int = 10,
    ):
        self.data_savepath = data_savepath
        self.data_network = data_network
        self.data_station = data_station
        self.data_location = data_location
        self.network_sampling_rate_banks_pooling = network_sampling_rate_banks_pooling
        self.ica_number = ica_number

    def _get_predictions(self, win_size, num_clusters):
        file_path = f'{self.data_savepath}clustering/{self.data_network}_{self.data_station}_{self.data_location}_{win_size}_{win_size}_{self.network_sampling_rate_banks_pooling}_ICA_{self.ica_number}_clusters_{num_clusters}.npz'
        return np.load(file_path)['predictions']

    def _build_crosstab_data(self,
                             cluster_1=3600,
                             num_clusters_1=10,
                             cluster_2=60,
                             num_clusters_2=10,
                             normalization=None):
        preds_1 = self._get_predictions(cluster_1, num_clusters_1)
        preds_2 = self._get_predictions(cluster_2, num_clusters_2)
        factor_difference = int(len(preds_2) / len(preds_1))

        print(
            f'Factor difference between predictions "{cluster_1}s window, {num_clusters_1} clusters" and "{cluster_2}s window, {num_clusters_2} clusters" : {factor_difference} \n'
        )

        preds_1_modified = np.repeat(preds_1, factor_difference)

        df_preds = pd.DataFrame({'predictions_1': preds_1_modified, 'predictions_2': preds_2})

        _normalization_options = ('all', 'index', 'columns')
        if normalization is None:
            ct_data = pd.crosstab(df_preds.predictions_1, df_preds.predictions_2)
        elif normalization in _normalization_options:
            ct_data = pd.crosstab(df_preds.predictions_1, df_preds.predictions_2, normalize=normalization)
            ct_data = ct_data * 100
        else:
            raise ValueError(
                f"Provided Normalization is not valid. This can be 'None' or [{_normalization_options}]. Kindly see https://pandas.pydata.org/docs/reference/api/pandas.crosstab.html for more info."
            )
        ct_data = ct_data.sort_index(ascending=False)
        return ct_data, factor_difference, fowlkes_mallows_score(df_preds.predictions_1, df_preds.predictions_2)

    def plot_clustering_crosstab(self,
                                 cluster_1=3600,
                                 num_clusters_1=10,
                                 cluster_2=60,
                                 num_clusters_2=10,
                                 normalization=None,
                                 **kwargs):

        ct_data, _, fws = self._build_crosstab_data(cluster_1, num_clusters_1, cluster_2, num_clusters_2, normalization)

        fig, ax = plt.subplots(figsize=(7, 7), **kwargs)
        _fmt = '.0f'
        if normalization is not None:
            _fmt = '.1f'
        s = sns.heatmap(ct_data, cmap='YlGnBu', annot=True, cbar=False, fmt=_fmt, linewidth=.5, ax=ax)
        _norm_title = ''
        if normalization is not None:
            for t in s.texts:
                t.set_text(t.get_text() + ' %')
            _norm_title = f'\nNormalization: {normalization}'
        ax.set_xlabel(f'{cluster_2}s window, {num_clusters_2} clusters')
        ax.set_ylabel(f'{cluster_1}s window, {num_clusters_1} clusters')
        ax.set_title('Comparison of Clusterings Predictions\nFWS similarity {:.2f}%{}'.format(fws * 100, _norm_title))
        ax.xaxis.set_ticks_position('top')
        ax.xaxis.set_label_position('top')

        _norm_fname = ''
        if normalization is not None:
            _norm_fname = f'_Normalization_{normalization}'
        plt.savefig(
            f'{self.data_savepath}figures/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_sampling_rate_banks_pooling}_ICA_{self.ica_number}_crosstab_{_norm_fname}_{cluster_1}_{num_clusters_1}_{cluster_2}_{num_clusters_2}.png'
        )

    def _custom_dendrogram_crosstab(self, linkage: np.array, ax: Axes, depth: int = 30, orientation='left', factor=1):
        # Show and get dendrogram
        with plt.rc_context({'lines.linewidth': 0.7}):
            dendrogram_infos = hierarchy.dendrogram(linkage,
                                                    p=depth,
                                                    truncate_mode='lastp',
                                                    color_threshold=0,
                                                    ax=ax,
                                                    orientation=orientation,
                                                    above_threshold_color='0.3',
                                                    count_sort=True,
                                                    labels=None,
                                                    show_leaf_counts=False)

            R = hierarchy.dendrogram(
                linkage,
                p=depth,
                truncate_mode='lastp',
                color_threshold=0,
                no_plot=True,
                orientation=orientation,
                above_threshold_color='0.3',
                count_sort=True,
                labels=None,
            )

        # Extract informations
        coordinates, _ = get_leaves(dendrogram_infos, ax)

        # leave population
        pops = [x for x in R['ivl']]

        # Plot leave nodes
        node_style = dict(ms=5, mec='0.3', mew=0.7, clip_on=False)
        for coordinate, color, pop in zip(coordinates, COLORS, pops):
            if orientation == 'left':
                ax.plot(0, coordinate, 'o', mfc=color, **node_style)
                index = int((coordinate - 5) / 10) + 1
                label = '{:d}\n{}\n{}'.format(index, pop,
                                              '[' + str(int(pop.replace('(', '').replace(')', '')) * factor) + ']')
                ax.text(-0.1, coordinate, label, color=color, va='center')
            if orientation == 'bottom':
                ax.plot(coordinate, 0, 'o', mfc=color, **node_style)
                index = int((coordinate - 5) / 10) + 1
                label = '{:d}\n{}'.format(index, pop)
                ax.text(coordinate, -0.1, label, color=color, ha='center')

    def _get_linkage(self, win_size):
        return np.load(
            f'{self.data_savepath}clustering/{self.data_network}_{self.data_station}_{self.data_location}_{win_size}_{win_size}_{self.network_sampling_rate_banks_pooling}_linkage.npy'
        )

    def plot_crosstab_dendrograms(self,
                                  cluster_1=3600,
                                  num_clusters_1=10,
                                  cluster_2=60,
                                  num_clusters_2=10,
                                  normalization=None,
                                  **kwargs):

        ct_data, factor, fws = self._build_crosstab_data(cluster_1, num_clusters_1, cluster_2, num_clusters_2,
                                                         normalization)

        f, ax = plt.subplots(2,
                             2,
                             gridspec_kw={
                                 'width_ratios': [1, 8],
                                 'height_ratios': [8, 1]
                             },
                             figsize=(8, 8),
                             **kwargs)
        _fmt = '.0f'
        if normalization is not None:
            _fmt = '.1f'
        s = sns.heatmap(ct_data, cmap='YlGnBu', annot=True, cbar=False, fmt=_fmt, linewidth=.5, ax=ax[0, 1])
        _norm_title = ''
        if normalization is not None:
            for t in s.texts:
                t.set_text(t.get_text() + ' %')
            _norm_title = f'\nNormalization: {normalization}'
        ax[0, 1].set_xlabel(f'{cluster_2}s window, {num_clusters_2} clusters')
        ax[0, 1].set_ylabel(f'{cluster_1}s window, {num_clusters_1} clusters')
        ax[0, 1].set_title('Comparison of Clusterings Predictions\nFWS similarity {:.2f}%{}'.format(
            fws * 100, _norm_title))
        ax[0, 1].xaxis.set_ticks_position('top')
        ax[0, 1].xaxis.set_label_position('top')
        ax[0, 1].yaxis.set_ticks_position('right')
        ax[0, 1].yaxis.set_label_position('right')

        linkage_1 = self._get_linkage(win_size=cluster_1)
        self._custom_dendrogram_crosstab(linkage=linkage_1, ax=ax[0, 0], depth=num_clusters_1, factor=factor)
        linkage_2 = self._get_linkage(win_size=cluster_2)
        self._custom_dendrogram_crosstab(linkage=linkage_2, ax=ax[1, 1], depth=num_clusters_2, orientation='bottom')

        ax[1, 0].set_axis_off()
        _norm_fname = ''
        if normalization is not None:
            _norm_fname = f'_Normalization_{normalization}'
        plt.savefig(
            f'{self.data_savepath}figures/{self.data_network}_{self.data_station}_{self.data_location}_{self.network_sampling_rate_banks_pooling}_ICA_{self.ica_number}_crosstab_dendrogram_{_norm_fname}_{cluster_1}_{num_clusters_1}_{cluster_2}_{num_clusters_2}.png'
        )
